
% IR_mes_21_beta11_E1.m

%%
% Impulse response measurement 
%     using TSP(Linear-SS), Log-SS, NW-SS, and MN-SS signals.
%     All swept sine signals are synthesized based on signal energy spectrum. 

%  Please check the parameter values, and set the switch values before use

%   23/May/ 2015 
%   by Yutaka KANEDA, Tokyo Denki UNiv.
% kaneda@c.dendai.ac.jp

% IR_mes_21_beta11 ->

%% ==============================
clear all; close all;
set(0,'DefaultAxesFontSize',10,'DefaultTextFontSize',10,'DefaultUicontrolFontSize',12)

%% ===== Parameters setting ===========
% --- Signal parameters -----
fs = 48000;       % Sampling frequency
NN = 2^16;        % Signal length (1 period). Lower limit = 2^15
FL = fs * 0.01;   % Frame length for noise spectrum estimation 
                  %  = 10ms -> 100Hz resolution.
Amp = 0.5;        % Signal amplitude

% --- Switches ------------------------
signal_type = 4;  %  1: TSP
                  %  2: Log-SS
                  %  3: NW-SS
                  %  4: MN-SS
audio_in_out =1;  % Audio in-out
                  %   1: wavplay & wavrecord 
                  %   2: eASIO driverf
                  %   3: Simulation

% --- Other setting
JJ = NN/2;        % Effective signal length
SL = FL/2;        % Shift length fo noise spectrum estimation
% --- 
kk = 1:NN/2+1;    % Discrete frequency number
tt = (0:NN-1)/fs;  % Time axis

%% ===== Measurement of noise spectrum PN =====
% for NW- and MN- signals, and 
% to display noise component in the measured response

figure(101);
    text(0.05,0.8,'***  Starting noise recording  ***','FontSize',16)
    text(0.05,0.7,'   Keep quiet for 2 seconds','FontSize',16); drawnow;
    drawnow;

switch audio_in_out
    case 1  % use wavrecord 
        bgn0 = wavrecord(fs*2, fs);  % 
    case 2  % useeASIO driverf
        bgn0 = pa_wavrecord( 1, 1, fs*2, fs,  [], 'asio' );
    case 3  % Simulation 
        bgn0 = wavread('test_Mukyo_noise_UA101_10s');
    otherwise
        disp('! audio_in_out: parameter error'); return
end

figure(101); text(0.05,0.6,'*** Recording has finished  ***','FontSize',16); drawnow;

bgn = bgn0;

% --- Remove very low frequency noise
load 'HPF70_30Hz_48k_80';     % High Pass Filter for removing very low frequency noise
                              % Comment out this section if it isnft needed
                              % FIR filter with coefficient hHPF70_48kh
bgn = conv(bgn, HPF70_48k);     % Filtering
bgn = bgn(length(HPF70_48k)/2+1:end-length(HPF70_48k)/2+1);
                              % Correct time axis

% ---  Mean power spectrum --------------------------
PN = mps21(bgn(fs+1:2*fs), FL, SL, NN)';
                                % Calculate mean power spectrum of noise using 1 sec data 
figure(1);
LPN = length(PN);
ff = (0:LPN/2)*fs/LPN;
semilogx(ff,10*log10(PN(1:LPN/2+1)));
hold on; grid on; xlabel('Frequency (Hz)'); ylabel('Power (dB)');
xlim([30,fs/2]); title('Noise power spectrum'); drawnow;

%% == Signal synthesis (Power spectrum -> Group delay -> Phase Characteristic) 
% --- Power spectrum of signal
switch signal_type
    case 1; SType=' TSP ';   PS = ones(1,length(kk));
    case 2; SType='LogSS';   PS = 1./kk; 
    case 3; SType='NW-SS';   PS = PN(kk);
    case 4; SType='MN-SS';   PS = sqrt(PN(kk)); 
end

% ff1 = (0:NN/2)/NN*fs;    
% figure(21); subplot(311); plot(ff1, PS); 
% title('Signal Power Spectrum'); grid on; hold on;

% --- Synthesize swept sine signal
k1 =0;     % k1: Group delay D(k1)=0
k2=NN/2;   % k2: Group delay D(k2)=JJ

ss = SSsyn01(PS, k1, k2, NN, JJ, Amp);  % Synthesize swept sine signal 

figure(22);
subplot(211);
plot(tt, ss); xlim([0,inf])
title(['Measurement signal (',SType,')']);
grid; xlabel('Time (s)'); ylabel('Relative Amplitude');
subplot(212);
specgram(ss, [], fs);
xlabel ('Time (s)'); ylabel ('Frequency (Hz)'); zlabel ('dB')
caxis([-80 20]); 
% colorbar

%% --- Play signal & record its response (or its simulation)
%          and impulse response calculation
% ----- Set excitation signal 
play_ss = [ss, ss];   % 

% --- Taper window at the startpoint
nww = fs/1000*10;        % taper length 10ms
ww = hanning(nww)';       % taper window
play_ss(1:nww/2) = play_ss(1:nww/2) .* ww(1:nww/2);
% play_ss(end-nww/2+1:end) = play_ss(end-nww/2+1:end) .* ww(end-nww/2+1:end);

%----------------- Impulse response measurement --------------------
figure(101);
    text(0.05,0.4,'*** Starting impulse response measurement ***','FontSize',16); drawnow;
    text(0.05,0.3,['   Keep quiet for ',num2str(ceil(length(play_ss)/fs)),' second'],'FontSize',16); drawnow;

switch audio_in_out
    case 1  % measurement with functions wavplay  wavrecord 
        wavplay(play_ss,fs,'async'); % play
        rec0 = wavrecord(length(play_ss),fs);  % recording
    case 2  % Measurement using ASIO driver
        rec0 = pa_wavplayrecord( play_ss', [], fs, length(play_ss), 1, 1, [], 'asio' );
    case 3  % Simulation
        imp = wavread('imp11');
        rec0 = conv(play_ss, imp)';
        rec0 = rec0 + bgn0(1:length(rec0));
    otherwise
        disp('! audio_in_out: parameter error'); return
end

figure(101); text(0.05,0.2,'***  Measurement has finished  ***','FontSize',16); drawnow;

% --- Remove very low frequency noise
load 'HPF70_30Hz_48k_80';    % High Pass Filter for removing very low frequency noise
                              % Comment out this section if it isnft needed
                              % FIR filter with coefficient hHPF70_48kh
rec0 = conv(rec0, HPF70_48k);     % Filtering
rec = rec0(length(HPF70_48k)/2+1:end-length(HPF70_48k)/2+1);
                              % Time axis correction
 
%% ----------------------------------------------
%   Waveform & spectrogram of recording signal
%     (Check recording signal level, noise level and nonlinear distortion.)
% ----------------------------------------------

figure(2)
plot( (1:length(rec))/fs ,rec)
title(['Recorded Signal Waveform (two periods + \alpha ) \newline ',...
'(Make sure clipping does not occur, and acceptable SNR)']); 
xlabel('Time (s)'); ylabel('Relative Amplitude'); xlim([0,length(rec)/fs]);
hold on;for ii =1:2;plot([ii*NN,ii*NN]/fs,[-1,1],'g--');end
ylim([-1,1])

figure(3)
[bb2,ff,tt1] = specgram(rec,[],fs);
max2 = max(max(20*log10(abs(bb2))));
imagesc(tt1,ff/1000,20*log10(abs(bb2)),[max2-100, max2]);
colorbar;axis xy;grid on
title(['Spectrogram of recorded signal \newline',...
    ' (Check harmonic distortion and irregular noise)']); 
xlabel('Time (s)'); ylabel('Frequency (kHz)')
             % Low level distortion, e. g. -80dB, can be disregarded 
% figure(4)
% [bb2,ff,tt1] = specgram(rec(NN+1:2*NN),256,fs);
% max2 = max(max(20*log10(abs(bb2))));
% imagesc(tt1,ff/1000,20*log10(abs(bb2)),[max2-100, max2]);
% colorbar;axis xy;grid on
% title(' Spectrogram of recording signal (one period)'); xlabel('Time (s)'); ylabel('Frequency (kHz)')

% --- extract the second period 

rec2 = cross_fade_01(rec, NN)';  % Cross-fade extraction of the second period 

figure(31);
plot(tt,rec2); grid on; title(['The second period of recorded signal \newline', ...
             '(Check overflow error)']);
xlabel('Time (s)'); ylabel('Relative Amplitude'); xlim([0,length(rec2)/fs]);
ylim([-1,1])

figure(32);
[bb2,ff,tt1] = specgram(rec2,512,fs);
max2 = max(max(20*log10(abs(bb2))));
imagesc(tt1,ff/1000,20*log10(abs(bb2)),[max2-100, max2]);
colorbar;axis xy;grid on
title(' Spectrogram of the second period'); 
xlabel('Time (s)'); ylabel('Frequency (kHz)')

%%---- Multiply the inverse of SS signal and derive the impulse response   ----------

REC = fft(rec2); % Frequency domain representation of the second period
IR = REC./ fft(ss);
ir = real(ifft(IR));

figure(41)
plot(tt,ir); title('Impulse Response'); grid on;
xlabel('Time (s)'); ylabel('Relative Amplitude'); 

figure(42)
specgram(ir, 512/8, fs); caxis([-110 -20]); colorbar;
title(' Impulse Response '); xlabel('Time'); ylabel('Frequency')

%%---- Noise component in the measured impulse response -----------------
% --- When noise only period is short (NN > length(bgn)=2*fs) noise is connected in roop.
Lw = 509;                  % cross-fade length 
                           % about 10ms for 48kHz sampling
ww = hanning(Lw);          % window
Lw2 = (Lw-1)/2;            % half-window length

bgn9 = bgn;                % temporaly variable
bgn9(1:Lw2) = bgn(1:Lw2).*ww(1:Lw2) + ...
    bgn(end-Lw2+1:end).*ww(end-Lw2+1:end);  % cross-fade left-end and right-end 
bgn9 = bgn9(1:end-Lw2);    % latest data is unnecessary 

Nrep = ceil(2*NN / length(bgn9)) -1; % Repeat times so as to have the same data length
bgn99 = bgn9;
for ii =1:Nrep; bgn99 = [bgn99; bgn9]; end % connection of bgn9

% --- Calculate noise component
bgn999 = cross_fade_01(bgn99, NN)';  % cross-fade extraction of the second period
BGN = fft(bgn999);
BGN = BGN./ fft(ss);
bgn_ir = real(ifft(BGN));

npowerdB = 10*log10(mean(bgn_ir.^2));
disp([SType, ': Noise level = ',num2str(npowerdB,'%6.1f'), 'dB'])

% ==============================================================

%---- Calculation of the frequency response from the impulse response  -----
% --- Extract the main part of th impulse response
% icut1 = round(20/1000*fs);       % Extract 20ms before the Maximum amplitude time (Temporal) 
% icut2 = round(600/1000*fs);     % Extract 600ms after the Maximum amplitude time (Temporal)
% [amax, idx] = max(ir);          % Find the maximum amplitude time
% ir3 = [ir, ir, ir]; 
% ir_main = ir3(NN+idx-icut1+1: NN+idx+icut2); % Extraction(+1 makes even signallength)

[ir_main, idx1] = extract_ir_01(ir, fs);

tt_main = (idx1: idx1+length(ir_main)-1)/fs;       % Variable for time axis

figure(43)
plot(tt_main,ir_main); title('Main part of impulse response')
xlabel('Time'); ylabel('Amplitude'); grid on;
figure(44)
specgram(ir_main, 64, fs); caxis([-110 -20]); colorbar;
title(' Main part of impulse response '); xlabel('Time (s)'); ylabel('Frequency (Hz)')

% --- Frequency response
HH1 = fft(ir_main,length(ir_main)*4);              
HH2 = abs(HH1).^2;

%---- Mean power spectrum of noise component  ----------------

PNS = mps21(bgn_ir, 2^11, 2^11/2,2^13*8); % To increase frequency resolution
PNS = PNS * length(ir_main);   % Noise level adjustment for the length of theimpulse response

%---  Frequency response  ---------

figure(5)
% Frequency response (measurement result)
ffHH2 = (0:length(HH2)/2)*fs/length(HH2);
semilogx(ffHH2,10*log10(HH2(1:length(HH2)/2+1)));
hold on; grid on;
% Noise 
ffN = (0:length(PNS)/2)*fs/length(PNS);
semilogx(ffN, 10*log10(PNS(1:length(PNS)/2+1)),'r');
title('Measurement result');
xlabel('Frequency (Hz)'); ylabel('Relative power(dB)');
legend('Measurement result','Noise component',  4);
xlim([30 25000]);

%% --- Save data 
% save 'PN_MN' PN;
% save 'ir_MN' ir;
% save 'ir_main_MN' ir_main;

return
